class CorpusAnalysisApp {
  constructor(corpusId) {
    this.data = {
      stopwords: undefined,
      originalStopwords: {},
      stopwordCache: {},
      promises: {getStopwords: undefined}
    };

    // HTML elements
    this.elements = {
      container: document.querySelector('#corpus-analysis-app-container'),
      extensionTabs: document.querySelector('#corpus-analysis-app-extension-tabs'),
      initModal: document.querySelector('#corpus-analysis-app-init-modal'),
      overview: document.querySelector('#corpus-analysis-app-overview')
    };
    // Materialize elements
    this.elements.m = {
      extensionTabs: M.Tabs.init(this.elements.extensionTabs),
      initModal: M.Modal.init(this.elements.initModal, {dismissible: false})
    };

    this.extensions = {};

    this.settings = {
      corpusId: corpusId
    };
  }

  getStopwords() {
    this.data.promises.getStopwords = new Promise((resolve, reject) => {
      Requests.corpora.entity.getStopwords()
        .then((response) => {
          response.json()
            .then((json) => {
              for (let [key, value] of Object.entries(json)) {
                this.data.originalStopwords[key] = value;
              }
              this.data.stopwords = json;
              resolve(this.data.stopwords);
            })
            .catch((error) => {
              reject(error);
            });
        });
    });
    return this.data.promises.getStopwords;
  }

  init() {
    this.disableActionElements();
    this.elements.m.initModal.open();
    // Init data
    this.data.cqiClient = new cqi.CQiClient('/cqi_over_sio', this.settings.corpusId);
    this.data.cqiClient.connect('anonymous', '')
      .then((cqiStatus) => {
        return this.data.cqiClient.corpora.list();
      })
      .then((cqiCorpora) => {
          this.data.corpus = {o: cqiCorpora[0]};
          console.log(this.data.corpus.o.staticData);
          this.renderGeneralCorpusInfo();
          this.renderTextInfoList();
          this.renderTextProportionsGraphic()
          this.renderFrequenciesGraphic();
          this.renderBoundsGraphic();
          // TODO: Don't do this hgere
          this.data.corpus.o.updateDb();
          this.enableActionElements();
          for (let extension of Object.values(this.extensions)) {extension.init();}
          this.elements.m.initModal.close();
        },
        (cqiError) => {
          let errorString = `${cqiError.code}: ${cqiError.constructor.name}`;
          let errorsElement = this.elements.initModal.querySelector('.errors');
          let progressElement = this.elements.initModal.querySelector('.progress');
          errorsElement.innerText = errorString;
          errorsElement.classList.remove('hide');
          app.flash(errorString, 'error');
          progressElement.classList.add('hide');
        }
      );

  
    // Add event listeners
    for (let extensionSelectorElement of this.elements.overview.querySelectorAll('.extension-selector')) {
      extensionSelectorElement.addEventListener('click', () => {
        this.elements.m.extensionTabs.select(extensionSelectorElement.dataset.target);
      });
    }

    let frequenciesStopwordSettingModal = document.querySelector('#frequencies-stopwords-setting-modal');
    let frequenciesStopwordSettingModalButton = document.querySelector('#frequencies-stopwords-setting-modal-button');
    frequenciesStopwordSettingModalButton.addEventListener('click', () => {
      this.data.stopwordCache = {};
      const stopwordsCopy = Object.assign({}, this.data.stopwords);
      for (let [key, value] of Object.entries(stopwordsCopy)) {
        this.data.stopwordCache[key] = value;
      }
      this.renderStopwordSettingsModal(this.data.stopwords);
      M.Modal.init(frequenciesStopwordSettingModal, {dismissible: false});
    });

    for (let actionButton of document.querySelectorAll('.frequencies-stopword-setting-modal-action-buttons')) {
      actionButton.addEventListener('click', (event) => {
        let action = event.target.closest('.frequencies-stopword-setting-modal-action-buttons').dataset.action;
        if (action === 'submit') {
          this.renderFrequenciesGraphic();
        } else if (action === 'cancel') {
          this.data.stopwords = this.data.stopwordCache;
          this.renderFrequenciesGraphic();
        }
      });
    }
  }

  registerExtension(extension) {
    if (extension.name in this.extensions) {
      console.error(`Can't register extension ${extension.name}: Already registered`);
      return;
    }
    this.extensions[extension.name] = extension;
    if ('cQiClient' in this.data && this.data.cQiClient.connected) {extension.init();}
  }

  disableActionElements() {
    let actionElements = this.elements.container.querySelectorAll('.corpus-analysis-action');
    for (let actionElement of actionElements) {
      if (actionElement.nodeName === 'INPUT') {
        actionElement.disabled = true;
      } else if (actionElement.nodeName === 'SELECT') {
        actionElement.parentNode.querySelector('input.select-dropdown').disabled = true;
      } else {
        actionElement.classList.add('disabled');
      }
    }
  }

  enableActionElements() {
    let actionElements = this.elements.container.querySelectorAll('.corpus-analysis-action');
    for (let actionElement of actionElements) {
      if (actionElement.nodeName === 'INPUT') {
        actionElement.disabled = false;
      } else if (actionElement.nodeName === 'SELECT') {
        actionElement.parentNode.querySelector('input.select-dropdown').disabled = false;
      } else {
        actionElement.classList.remove('disabled');
      }
    }
  }

  renderGeneralCorpusInfo() {
    let corpusData = this.data.corpus.o.staticData;
    document.querySelector('.corpus-num-tokens').innerHTML = corpusData.corpus.counts.token;
    document.querySelector('.corpus-num-s').innerHTML = corpusData.corpus.counts.s;
    document.querySelector('.corpus-num-unique-words').innerHTML = Object.entries(corpusData.corpus.freqs.word).length;
    document.querySelector('.corpus-num-unique-lemmas').innerHTML = Object.entries(corpusData.corpus.freqs.lemma).length;
    document.querySelector('.corpus-num-unique-pos').innerHTML = Object.entries(corpusData.corpus.freqs.pos).length;
    document.querySelector('.corpus-num-unique-simple-pos').innerHTML = Object.entries(corpusData.corpus.freqs.simple_pos).length;
  }

  renderTextInfoList() {
    let corpusData = this.data.corpus.o.staticData;
    let corpusTextInfoListElement = document.querySelector('.corpus-text-info-list');
    let corpusTextInfoList = new CorpusTextInfoList(corpusTextInfoListElement);
    let texts = corpusData.s_attrs.text.lexicon;
    let textData = [];
    for (let i = 0; i < Object.entries(texts).length; i++) {
      let resource = {
        title: corpusData.values.s_attrs.text[i].title,
        publishing_year: corpusData.values.s_attrs.text[i].publishing_year,
        num_tokens: corpusData.s_attrs.text.lexicon[i].counts.token,
        num_sentences: corpusData.s_attrs.text.lexicon[i].counts.s,
        num_unique_words: Object.entries(corpusData.s_attrs.text.lexicon[i].freqs.word).length,
        num_unique_lemmas: Object.entries(corpusData.s_attrs.text.lexicon[i].freqs.lemma).length,
        num_unique_pos: Object.entries(corpusData.s_attrs.text.lexicon[i].freqs.pos).length,
        num_unique_simple_pos: Object.entries(corpusData.s_attrs.text.lexicon[i].freqs.simple_pos).length
      };
  
      textData.push(resource);
    }

    corpusTextInfoList.add(textData);

    let textCountChipElement = document.querySelector('.text-count-chip');
    textCountChipElement.innerHTML = `Text count: ${corpusData.corpus.counts.text}`;
  }

  renderTextProportionsGraphic() {
    let corpusData = this.data.corpus.o.staticData;
    let textProportionsGraphicElement = document.querySelector('#text-proportions-graphic');
    let texts = Object.entries(corpusData.s_attrs.text.lexicon);
    let graphData = [
      {
        values: texts.map(text => text[1].counts.token),
        labels: texts.map(text => `${corpusData.values.s_attrs.text[text[0]].title} (${corpusData.values.s_attrs.text[text[0]].publishing_year})`),
        type: 'pie'
      }
    ];
    let graphLayout = {
      showlegend: true,
      height: 486,
      margin: {
        l: 10,
        r: 10,
        b: 10,
        t: 10
      },
      legend: {
        "orientation": "h",
        font: {
          size: 10
        }
      }
    };
    let config = {
      responsive: true,
      displaylogo: false
    };

    Plotly.newPlot(textProportionsGraphicElement, graphData, graphLayout, config);
  }

  async renderFrequenciesGraphic() {
    let corpusData = this.data.corpus.o.staticData;
    let frequenciesTokenCategoryDropdownElement = document.querySelector('[data-target="frequencies-token-category-dropdown"]');
    let frequenciesTokenCategoryDropdownListElement = document.querySelector("#frequencies-token-category-dropdown");
    let frequenciesGraphicElement = document.querySelector('#frequencies-graphic');
    let texts = Object.entries(corpusData.s_attrs.text.lexicon);
    let graphtype = document.querySelector('.frequencies-graph-mode-button.disabled').dataset.graphType;
    let graphModeButtons = document.querySelectorAll('.frequencies-graph-mode-button');

    frequenciesTokenCategoryDropdownListElement.addEventListener('click', (event) => {
      frequenciesTokenCategoryDropdownElement.firstChild.textContent = event.target.innerHTML;
      this.renderFrequenciesGraphic(corpusData);
    });

    graphModeButtons.forEach(graphModeButton => {
      graphModeButton.addEventListener('click', (event) => {
        graphModeButtons.forEach(btn => {
          btn.classList.remove('disabled');
        });
        event.target.closest('.frequencies-graph-mode-button').classList.add('disabled');
        this.renderFrequenciesGraphic(corpusData);
      });
    });
    
    let tokenCategory = frequenciesTokenCategoryDropdownElement.firstChild.textContent.toLowerCase();
    
    let graphData = await this.createFrequenciesGraphData(tokenCategory, texts, corpusData, graphtype);
    let graphLayout = {
      barmode: graphtype === 'bar' ? 'stack' : '',
      margin: {
        t: 20,
        l: 50
      },
      yaxis: {
        showticklabels: graphtype === 'markers' ? false : true
      },
    };
    let config = {
      responsive: true,
      modeBarButtonsToRemove: ['zoom2d', 'select2d', 'lasso2d', 'zoomIn2d', 'zoomOut2d', 'autoScale2d', 'resetScale2d'],
      displaylogo: false
    };
    Plotly.newPlot(frequenciesGraphicElement, graphData, graphLayout, config);
  }

  async createFrequenciesGraphData(category, texts, corpusData, graphtype) {
    let stopwords = this.data.stopwords;
    if (this.data.stopwords === undefined) {
      stopwords = await this.getStopwords();
    }
    let stopwordList = [];
    Object.values(stopwords).forEach(stopwordItems => {
      stopwordItems.forEach(stopword => {
        stopwordList.push(stopword);
      });
    });

    let graphData = [];
    let filteredData = Object.entries(corpusData.corpus.freqs[category])
      .sort((a, b) => b[1] - a[1])
      .filter(item => !stopwordList.includes(corpusData.values.p_attrs[category][item[0]].toLowerCase()))
      .slice(0, 5);

    if (graphtype !== 'markers') {
      for (let item of filteredData) {
        let data = {
          x: texts.map(text => `${corpusData.values.s_attrs.text[text[0]].title} (${corpusData.values.s_attrs.text[text[0]].publishing_year})`),
          y: texts.map(text => text[1].freqs[category][item[0]] || 0),
          name: corpusData.values.p_attrs[category][item[0]],
          type: graphtype
        };
        graphData.push(data);
      }
    } else {
      for (let item of filteredData) {
        let size = texts.map(text => text[1].freqs[category][item[0]] || 0); 
        let data = {
          x: texts.map(text => `${corpusData.values.s_attrs.text[text[0]].title} (${corpusData.values.s_attrs.text[text[0]].publishing_year})`),
          y: texts.map(text => corpusData.values.p_attrs[category][item[0]]),
          name: corpusData.values.p_attrs[category][item[0]],
          text: texts.map(text => `${corpusData.values.p_attrs[category][item[0]]}<br>${text[1].freqs[category][item[0]] || 0}`),
          mode: 'markers',
          marker: {
            size: size,
            sizeref: 0.4
          }
        };
        graphData.push(data);
      }
    }
    return graphData;
  }

  renderStopwordSettingsModal(stopwords) {
    let stopwordInputField = document.querySelector('#stopword-input-field');
    let userStopwordListContainer = document.querySelector('#user-stopword-list-container');
    let stopwordLanguageSelection = document.querySelector('#stopword-language-selection');
    let stopwordLanguageChipList = document.querySelector('#stopword-language-chip-list');
    let deleteLanguageStopwordListEntriesButton = document.querySelector('#delete-language-stopword-list-entries-button');
    let resetLanguageStopwordListEntriesButton = document.querySelector('#reset-language-stopword-list-entries-button');

    stopwordLanguageChipList.innerHTML = '';
    userStopwordListContainer.innerHTML = '';
    
    // Render stopword language selection. Set english as default language. Filter out user_stopwords.
    for (let language of Object.keys(stopwords)) {
      if (language !== 'user_stopwords') {
        if (language === 'english') {
          let optionElement = Utils.HTMLToElement(`<option value="${language}" selected>${language}</option>`);
          stopwordLanguageSelection.appendChild(optionElement);
        } else {
          let optionElement = Utils.HTMLToElement(`<option value="${language}">${language}</option>`);
          stopwordLanguageSelection.appendChild(optionElement);
        }
      }
    }

    // Render user stopwords over input field.
    if (this.data.stopwords['user_stopwords'].length > 0) {
      for (let word of this.data.stopwords['user_stopwords']) {
        let chipElement = Utils.HTMLToElement(`<div class="chip">${word}<i class="close material-icons">close</i></div>`);
        chipElement.addEventListener('click', (event) => {
          let removedListItem = event.target.closest('.chip').firstChild.textContent;
          this.data.stopwords['user_stopwords'] = this.data.stopwords['user_stopwords'].filter(item => item !== removedListItem);
        });
        userStopwordListContainer.appendChild(chipElement);
      }
    }

    // Render english stopwords as default ...
    this.renderStopwordLanguageChipList('english', stopwords['english']);
    
    // ... or render selected language stopwords.
    stopwordLanguageSelection.addEventListener('change', (event) => {
      this.renderStopwordLanguageChipList(event.target.value, stopwords[event.target.value]);
    });

    // Eventlistener for deleting all stopwords of a language.
    deleteLanguageStopwordListEntriesButton.addEventListener('click', (event) => {
      let selectedLanguage = stopwordLanguageSelection.value;
      this.data.stopwords[selectedLanguage] = [];
      stopwordLanguageChipList.innerHTML = '';
      this.buttonRendering();
    });

    // Eventlistener for resetting all stopwords of a language to the original stopwords.
    resetLanguageStopwordListEntriesButton.addEventListener('click', () => {
      let selectedLanguage = stopwordLanguageSelection.value;
      this.data.stopwords[selectedLanguage] = this.data.originalStopwords[selectedLanguage];
      this.renderStopwordLanguageChipList(selectedLanguage, this.data.stopwords[selectedLanguage]);
    });

    // Initialize Materialize components.
    M.Chips.init(
        stopwordInputField, 
        {
          placeholder: 'Add stopwords',
          onChipAdd: (event) => {
            let userStopwords = [];
            for (let word of event[0].M_Chips.chipsData) {
              if (!this.data.stopwords['user_stopwords'].includes(word.tag.toLowerCase())) {
                userStopwords.push(word.tag.toLowerCase());
              }
            }
            this.data.stopwords['user_stopwords'] = this.data.stopwords['user_stopwords'].concat(userStopwords);
          }
        }
    );
    M.FormSelect.init(stopwordLanguageSelection);

  }

  buttonRendering() {
    let stopwordLanguageSelection = document.querySelector('#stopword-language-selection');
    let deleteLanguageStopwordListEntriesButton = document.querySelector('#delete-language-stopword-list-entries-button');
    let resetLanguageStopwordListEntriesButton = document.querySelector('#reset-language-stopword-list-entries-button');

    let selectedLanguage = stopwordLanguageSelection.value;
    let stopwordLength = this.data.stopwords[selectedLanguage].length;
    let originalStopwordListLength = this.data.originalStopwords[selectedLanguage].length;

    resetLanguageStopwordListEntriesButton.classList.toggle('blue', stopwordLength !== originalStopwordListLength);
    deleteLanguageStopwordListEntriesButton.classList.toggle('red', stopwordLength > 0);
    resetLanguageStopwordListEntriesButton.style.cursor = stopwordLength !== originalStopwordListLength ? 'pointer' : 'default';
    deleteLanguageStopwordListEntriesButton.style.cursor = stopwordLength > 0 ? 'pointer' : 'default';
  }
  
  renderStopwordLanguageChipList(language, stopwords) {
    let stopwordLanguageChipList = document.querySelector('#stopword-language-chip-list');
    stopwordLanguageChipList.innerHTML = '';
    for (let word of stopwords) {
      let chipElement = Utils.HTMLToElement(`<div class="chip">${word}<i class="close material-icons">close</i></div>`);
      chipElement.addEventListener('click', (event) => {
        let removedListItem = event.target.closest('.chip').firstChild.textContent;
        this.data.stopwords[language] = this.data.stopwords[language].filter(item => item !== removedListItem);
        this.buttonRendering();
      });
      stopwordLanguageChipList.appendChild(chipElement);
    }
    this.buttonRendering();
  }
  
  renderBoundsGraphic() {
    let corpusData = this.data.corpus.o.staticData;
    let boundsGraphicElement = document.querySelector('#bounds-graphic');

    let graphData = [];
    let texts = Object.entries(corpusData.s_attrs.text.lexicon);

    graphData = [{
      type: 'bar',
      x: texts.map(text => text[1].bounds[1] - text[1].bounds[0]),
      y: texts.map(text => corpusData.values.s_attrs.text[text[0]].title),
      base: texts.map(text => text[1].bounds[0]),
      text: texts.map(text => `${corpusData.values.s_attrs.text[text[0]].title} (${corpusData.values.s_attrs.text[text[0]].publishing_year})`),
      orientation: 'h',
      hovertemplate: '%{base} - %{x} <br>%{y}',
      showlegend: false
    }];

    let graphLayout = {
      barmode: 'stack',
      type: 'bar',
      showgrid: false,
      xaxis: {
        rangemode: 'nonnegative',
        autorange: true
      },
      yaxis: {
        autorange: true,
        showticklabels: false
      }
    };

    let config = {
      responsive: true,
      modeBarButtonsToRemove: ['zoom2d', 'select2d', 'lasso2d', 'zoomIn2d', 'zoomOut2d', 'autoScale2d', 'resetScale2d'],
      displaylogo: false
    };
  
    Plotly.newPlot(boundsGraphicElement, graphData, graphLayout, config);
  }
}
